#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id: method_doutput.py 915 2010-10-01 15:51:17Z smrcka $
# Due to analysis of clock domain crossing performed on one module only, the
# algorithm implemented in this python module does not support the application
# of the method of extending every input port of a critical signal path
# on a hardware design containing more than one modules sensitive on more than
# 2 clock signals.
'extend clock domain outputs'

import common
import model as m
from math import log, ceil


class DelayedInputModule(object):
    'delayed input module SMV implementation'
    name = 'delayed_input'
    def __str__(self):
        return '''module %s(signal, delayed)
{
    input signal : boolean;
    output delayed : boolean;

    init(delayed) := signal;
    if (signal = next(signal))
        next(delayed) := signal;
    else
        next(delayed) := {0, 1};
}''' % self.name

class DelayedInputModInst(object):
    'delayed input module instance statement'
    def __init__(self, input, output):
        self.name = common.uniqname('delayed_input')
        self.input = input
        self.output = output
    
    def __str__(self):
        return '%s : %s(%s, %s);' % (self.name, DelayedInputModule.name, \
            self.input, self.output)

def extend(model):
    "apply the method of extending all critical source ports"
    # flatten every "renaming module", i.e., every module containing the 
    # simplest assignment statements only 
    common.flatten_renaming_modules(model)

    # let only one module being sensed by more than one clock signals
    common.flatten_all_asyn_modules(model)

    # transform every assignment statement representing combinational logic to
    # structural description
    common.comb_logic_to_structural(model)
    common.debug2('Module clocks')
    common.print_found_clocks(model, 3)

    # remove a:=b; 
    common.remove_aliases(model)

    # find a module to extend
    model.asyn_module = common.find_module_to_extend(model)
    if not model.asyn_module:
        return None

    # find all instantiated modules and set them to appropriate clock domain
    common.debug1('Partitioning submodules of "%s" to clock domains' % \
        model.asyn_module.name)
    common.partition_clock_domain(model)
    common.print_clock_domains(model)

    # get the list of all module instances
    model.gates = common.get_modinst_list(model)

    common.debug1('Extending clock domain outputs (doutput method)')

    # make soruce and destination direct data flow
    common.make_src_dst_flow(model)

    # make transitive closure of both flow sets
    common.debug1('Searching for critical signal paths')
    if common.OPTS['verbose'] >= 4:
        common.debug4('0. iteration of source flow:')
        common.print_src_flow(model)
    iter = common.flow_transitive_closure(model.src_flow, model.iterable_src)
    common.print_src_flow(model)
    common.debug4('Source flow iterations: %d' % iter)
    if common.OPTS['verbose'] >= 4:
        common.debug4('0. iteration of destination flow:')
        common.print_dst_flow(model)
    iter = common.flow_transitive_closure(model.dst_flow, model.iterable_dst)
    common.debug4('Destination flow iterations: %d' % iter)
    common.get_critical_ports(model)
    common.print_dst_flow(model)

    # extend all critical source ports
    extend_critical_source_ports(model)
    common.debug1('Critical source ports: %d' % \
        len(model.critical_source_ports))
    common.debug2('%40s | %s' % ('gate', 'connected input signal'))
    common.debug2('-'*76)
    for gate, signal in model.critical_source_ports:
        common.debug2('%40s | %s' % (gate.name, signal))
    
    model.new_bits = len(model.critical_source_ports)
    common.debug1('Introduced new %d bits of state variables' % model.new_bits)

    return model

def extend_gate_input(model, gate, paramname, signal):
    '''extends the signal path ending in `modinst`-`input` wrt clock domain
    of `clock` with delayed input modules'''
    if not gate.param_map[paramname].type.compatible(m.BOOLEAN):
        raise NotImplementedError( \
          'unable to extend non-boolean type input port `%s` of module `%s`' \
            % (paramname, modinst.module.name))
    # create new signal connecting input-delay with gate
    d_var = m.Variable(common.uniqname('delay_%s' % signal.name), m.BOOLEAN)
    d_decl = m.VariableDeclStmt(d_var)
    model.asyn_module.stmts.insert(0, d_decl)
    # connect it to input port
    gate.param_map[paramname] = d_var
    # create input-delay
    di_inst = DelayedInputModInst(signal, d_var)
    model.asyn_module.stmts.append(di_inst)

def extend_critical_source_ports(model):
    '''extends every connection of model.critical_ports to a gate with a new
    instance of input-delay'''
    model.critical_source_ports = []
    for gate, signal in model.critical_ports:
        if signal not in model.critical_signals:
            continue
        # find name of an input port
        for paramname, connected in gate.param_map.items():
            if signal in common.get_vars(connected):
                extend_gate_input(model, gate, paramname, signal)
                model.critical_source_ports.append((gate, signal))
    if model.critical_source_ports:
        model.stmts.append(DelayedInputModule())
